# run_rbn_control_kernels.py
#
# Bryan Daniels
# 5.15.2020 branched from run_control_kernels.py
#
# Find control kernels for each attractor for a number of
# random networks created with RBN_Kim.  The work index
# is set by command line argument, which sets N, average
# degree, and proportion of inhibitory links according to
# a preset list.
#

import sys
from toolbox.simplePickle import save
from neet.boolean.randomnet import RBN_Kim,RBN,np
from control_kernel_analysis import ck_analysis,sampled_ck_analysis,dn_analysis

if len(sys.argv) != 2:
    print("Usage: python run_rbn_control_kernels.py [workIndex]")
    exit()

# () set static and varying parameters and select which to do using work index

phenotype = 'all' # 'internal'
phenotype_only = False # False (runs on and before 8/12/19 have effective value False)
require_inputs = True # False (runs before 5/1/20 have effective value False)
run_sampling_analysis = False # True
run_dn_analysis = True
thresholdNetwork = False # True (runs before 12/15/20 have effective value True)

Nlist = [10,15,20]
workIndex = int(sys.argv[1])
if thresholdNetwork:
    degList = [1,2,3,4,5]
    inhList = [0.1,0.3,0.5,0.7,0.9]
    thresholdParam = None #0.5 # (runs before 12/9/20 have effective value None)
    paramList = [ (N,deg,inh) for N in Nlist for deg in degList for inh in inhList ]
    N,deg,inh = paramList[workIndex]
else: # "standard" RBN
    N = Nlist[workIndex % 3]
    deg = 2
    p = 0.75 #0.25 #0.5
    self_loops = False # True (runs before 12/21/20 have effective value True)

# () set sampled control kernel parameters

sampleseed = 130
samples = 10000 #10000 #1000

# () set up a random network as net

seed = 123*workIndex
if thresholdNetwork:
    net = RBN_Kim(N,deg,inh,seed=seed,thresholdParam=thresholdParam)
    netname = "random_{}_tp{}_seed{}".format(workIndex,thresholdParam,seed)
else:
    np.random.seed(seed)
    net = RBN(N,deg,p,self_loops=self_loops)
    netname = "random_{}_N{}_k{}_p{}_seed{}".format(workIndex,N,deg,p,seed)

data = {}

if run_dn_analysis:
    # () run distinguishing node analysis

    data_dn = dn_analysis(net,
                          require_inputs=require_inputs)

    # () save output data to pickled file

    data.update(data_dn)
    data["name"] = netname
    data["net"] = net
    if thresholdNetwork:
        data.update({"mean degree":deg,
         "inhibitory probability":inh,
         "threshold parameter":thresholdParam,
         "network seed":seed,
        })
    else:
        data.update({"degree":deg,
         "bias":p,
         "network seed":seed,
        })
    filename = "control_kernel_{}.dat".format(netname)
    print("Outputting distinguishing node data to {}".format(filename))
    print()
    save(data,filename)

# () run control kernel analysis

data_ck = ck_analysis(net)

# () save exact control kernel output data to pickled file

data.update(data_ck)
data["name"] = netname
data["net"] = net
if thresholdNetwork:
    data.update({"mean degree":deg,
     "inhibitory probability":inh,
     "threshold parameter":thresholdParam,
     "network seed":seed,
    })
else:
    data.update({"degree":deg,
     "bias":p,
     "network seed":seed,
    })
filename = "control_kernel_{}.dat".format(netname)
print("Outputting data to {}".format(filename))
save(data,filename)

if run_sampling_analysis:
    # () run sampled control kernel analysis

    data_sampled = sampled_ck_analysis(net,seed=sampleseed,samples=samples,
                               phenotype=phenotype,phenotype_only=phenotype_only,
                               require_inputs=require_inputs)

    # () save sampled control kernel output data to pickled file

    data_sampled["name"] = netname
    data_sampled["net"] = net
    if thresholdNetwork:
        data_sampled.update({"mean degree":deg,
         "inhibitory probability":inh,
         "threshold parameter":thresholdParam,
         "network seed":seed,
        })
    else:
        data_sampled.update({"degree":deg,
         "bias":p,
         "network seed":seed,
         "self loops":self_loops,
        })
    filename_sampled = "sampled_control_kernel_{}.dat".format(netname)
    print("Outputting data_sampled to {}".format(filename_sampled))
    save(data_sampled,filename_sampled)

